#include "RdmaRepEndpoint.hpp"

RdmaRepEndpoint::RdmaRepEndpoint(struct rdma_cm_id *id, int sendQueueSize, int recvQueueSize,
                                 int sendMsgSize, int recvMsgSize, rocksdb::DB *db)
    : _cm_id(id), _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize), _sendMsgSize(sendMsgSize),
      _recvMsgSize(recvMsgSize), _db(db)
{
}

void RdmaRepEndpoint::processSendCompletion(struct ibv_wc *data)
{
    /*means data has been send to other side we can use this buffer*/
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    _sendBuffers.push((char *)data->wr_id);
}

int RdmaRepEndpoint::sendMessage(const char *buffer, uint32_t size)
{
    if (size > (uint32_t)_sendMsgSize)
    {
        std::ostringstream ss;
        ss << "Large Message size " << size;
        ss << " send buffer size " << _sendMsgSize;
        CPPLog::LOG_ERROR(ss);
        return -1;
    }
    char *sendBuffer = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return -1;
    sendBuffer = (char *)_sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    memcpy((void *)sendBuffer, buffer, size);
    int ret = rdma_post_send(_cm_id, (void *)sendBuffer, (void *)sendBuffer, size, _sendMr, 0);
    if (ret == -1)
    {
        std::ostringstream ss;
        ss << "rdma_post_send error occured errno " << errno << strerror(errno);
        CPPLog::LOG_ERROR(ss);
        return ret;
    }
    return ret;
}

void RdmaRepEndpoint::processRecvCompletion(struct ibv_wc *data)
{
    char *request = new char[data->byte_len];
    memcpy(request, (void *)data->wr_id, data->byte_len);
    struct MessageHeader *req = (struct MessageHeader *)request;
    rdma_post_recv(_cm_id, (void *)data->wr_id, (void *)data->wr_id, _recvMsgSize, _recvMr);
    if (req->type == MessageType::DELETE)
        processDelete(req);
    if (req->type == MessageType::GET)
        processGet(req);
    if (req->type == MessageType::PUT)
        processPut(req);
    delete[] request;
}
/*since RepEndpoint need not to wait for response from other servers etc we will send the response immediately*/
void RdmaRepEndpoint::processDelete(struct MessageHeader *req)
{
    rocksdb::Status s = _db->Delete(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize});
    void *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
        return;
    sendBuf = _sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->type = MessageType::SUCCESS;
    }
    int ret = rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize, _sendMr, 0);
    if (ret == -1)
    {
        std::ostringstream ss;
        ss << "rdma_post_send error occured errno " << errno << strerror(errno);
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaRepEndpoint::processGet(struct MessageHeader *req)
{
    char *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
    {
        return;
    }
    sendBuf = (char *)_sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    std::string value;
    rocksdb::Status s = _db->Get(rocksdb::ReadOptions(), {(char *)req + MessageHeaderSize, req->keySize}, &value);
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
    {
        response->type = MessageType::SUCCESS;
        response->valueSize = value.size();
        memcpy(response + MessageHeaderSize, value.c_str(), value.size());
    }
    int ret = rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize + value.size(), _sendMr, 0);
    if (ret == -1)
    {
        std::ostringstream ss;
        ss << "rdma_post_send error occured errno " << errno << strerror(errno);
        CPPLog::LOG_ERROR(ss);
    }
}

void RdmaRepEndpoint::processPut(struct MessageHeader *req)
{
    rocksdb::Status s = _db->Put(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize},
                                 {(char *)req + MessageHeaderSize + req->keySize, req->valueSize});
    char *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(_sendBuffersM);
    if (_sendBuffers.size() == 0)
    {
        CPPLog::LOG_ERROR("No send Buffer");
        return;
    }
    sendBuf = (char *)_sendBuffers.front();
    _sendBuffers.pop();
    lock.unlock();
    MessageHeader *response = (MessageHeader *)sendBuf;
    /*This id done to avoid else case*/
    response->type = MessageType::FAILURE;
    response->id = req->id;
    if (s.ok())
        response->type = MessageType::SUCCESS;
    int ret = rdma_post_send(_cm_id, sendBuf, sendBuf, MessageHeaderSize, _sendMr, 0);
    if (ret == -1)
    {
        std::ostringstream ss;
        ss << "rdma_post_send error occured errno " << errno << strerror(errno);
        CPPLog::LOG_ERROR(ss);
    }
}
