#include "TaskThread.hpp"

TaskThread::TaskThread(int id, int cpu, ConcurrentQueue *taskqueue,
                       std::unordered_map<uint32_t, RdmaRepEndpoint *> *clientRepMap,
                       std::unordered_map<uint32_t, RdmaRepEndpoint *> *serverRepMap,
                       std::unordered_map<uint32_t, RdmaSalEndpoint *> *salMap,
                       std::unordered_map<uint32_t, std::shared_ptr<CQEventData>> &cqEvents,
                       std::mutex &cqEventsMutex)
    : _id(id), _clientRepMap(clientRepMap), _serverRepMap(serverRepMap), _salMap(salMap), _cqEvents(cqEvents), _cqEventsMutex(cqEventsMutex)
{
    _taskQueue = taskqueue;
    if (pthread_create(&thread, NULL, &TaskThread::run, this))
    {
        CPPLog::LOG_ERROR("pthread create has been failed while creating taskthread");
        exit(0);
    }
    /*cpu_set_t cpuset;
    CPU_ZERO(&cpuset);
    CPU_SET(cpu, &cpuset);
    std::ostringstream ss;
    ss << "New Thread Setting CPU affinty " << cpu;
    CPPLog::LOG_ALWAYS(ss);
    if (pthread_setaffinity_np(thread, sizeof(cpu_set_t), &cpuset) != 0)
    {
        CPPLog::LOG_ERROR("Error calling pthread_setaffinity_np ");
    }*/
    pthread_setname_np(thread, "TaskThread");
}

TaskThread::~TaskThread()
{
    CPPLog::LOG_INFO("TaskThread Destructed");
    stop();
}

void TaskThread::stop()
{
    _stop = true;
    if (pthread_join(thread, NULL) == 0)
    {
        CPPLog::LOG_ERROR("pthread join failed");
    }
}

inline void *TaskThread::run(void *object)
{
    /*
     * This is because data from pthread_create is passed as void*
     */
    TaskThread *thread = reinterpret_cast<TaskThread *>(object);
    std::ostringstream ss;
    ss << "Running task thread" << thread->_id;
    CPPLog::LOG_ALWAYS(ss);
    while (!thread->_stop)
    {
        struct ibv_wc *data = NULL;
        // std::cout << "Get start\n";
        data = thread->_taskQueue->try_pop();
        if (data != NULL)
        {
            // std::cout << "TaskThread:: got data";
            if (data->status != IBV_WC_SUCCESS)
            {
                std::ostringstream ss;
                ss << "TaskThread : failed work completion : ";
                ss << ibv_wc_status_str(data->status) << " on qp " << data->qp_num;
                CPPLog::LOG_ERROR(ss);
                continue;
            }
            RdmaSalEndpoint *salEp = nullptr;
            RdmaRepEndpoint *repEp = nullptr;
            auto it = thread->_salMap->find(data->qp_num);
            if (it != thread->_salMap->end())
            {
                salEp = it->second;
            }
            // Now we have to check whether event came for replication group or not
            auto it2 = thread->_clientRepMap->find(data->qp_num);
            if (it2 != thread->_clientRepMap->end())
            {
                repEp = it2->second;
            }
            else
            {
                auto it3 = thread->_serverRepMap->find(data->qp_num);
                if (it3 != thread->_serverRepMap->end())
                    repEp = it3->second;
            }
            if (salEp == nullptr && repEp == nullptr)
            {
                std::ostringstream ss;
                ss << "RdmaSal : endpoint not registered for qp" << data->qp_num;
                CPPLog::LOG_INFO(ss);

                continue;
            }
            else if (salEp != nullptr)
                thread->processSalEvent(salEp, data);
            else if (repEp != nullptr)
                thread->processRepEvent(repEp, data);
            thread->_taskQueue->removeFromSet(data);
            delete data;
        }
    }
    return NULL;
}

void TaskThread::processSalEvent(RdmaSalEndpoint *ep, struct ibv_wc *data)
{
    std::cout << "processing sal event" << data->opcode << "\n";
    /* sal Request*/
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        ep->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(ep->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       ep->_recvMsgSize, ep->_recvMr);
        struct MessageHeader *req = (struct MessageHeader *)buffer;
        std::cout << "recv" << req->type << "\n";
        std::shared_ptr<CQEventData> cqevent = std::make_shared<CQEventData>(_serverRepMap->size() + _clientRepMap->size(), ep);
        uint32_t id = req->id;
        /*
         * We need to wait for replication and invalidation to be done for delete and put request,hence we are putting data in map
         */
        rocksdb::Status s;
        if (req->type == MessageType::GET)
        {
            ep->processGet(req);
        }
        else if (req->type == MessageType::DELETE)
        {
            std::unique_lock<std::mutex> lock(_cqEventsMutex);
            _cqEvents[id] = cqevent;
            lock.unlock();
            replicateSalRequest(buffer, data->byte_len);
            sendInvalidation(buffer);
            s = ep->_db->Delete(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize});
            if (s.ok())
                processReplicationDone(req->id, MessageType::SUCCESS, true);
            else
                processReplicationDone(req->id, MessageType::FAILURE, true);
        }
        else if (req->type == MessageType::PUT)
        {
            std::unique_lock<std::mutex> lock(_cqEventsMutex);
            _cqEvents[id] = cqevent;
            lock.unlock();
            replicateSalRequest(buffer, data->byte_len);
            sendInvalidation(buffer);
            s = ep->_db->Put(rocksdb::WriteOptions(), {(char *)req + MessageHeaderSize, req->keySize},
                             {(char *)req + MessageHeaderSize + req->keySize, req->valueSize});
            if (s.ok())
                processReplicationDone(req->id, MessageType::SUCCESS, true);
            else
                processReplicationDone(req->id, MessageType::FAILURE, true);
        }
        else
        {
            std::ostringstream ss;
            ss << "SalRequest invalid req type" << data->opcode;
            CPPLog::LOG_ERROR(ss);
        }
        delete[] buffer;
    }
    break;
    default:
        std::ostringstream ss;
        ss << "TaskThread default opcode : " << data->opcode;
        CPPLog::LOG_INFO(ss);
        break;
    }
}

void TaskThread::processRepEvent(RdmaRepEndpoint *ep, struct ibv_wc *data)
{
    std::cout << "processing rep event" << data->opcode << "\n";
    switch (data->opcode)
    {
    case IBV_WC_SEND:
        ep->processSendCompletion(data);
        break;
    case IBV_WC_RECV:
    {
        // it->second->processRecvCompletion(data);
        char *buffer = new char[data->byte_len];
        memcpy(buffer, (void *)data->wr_id, data->byte_len);
        rdma_post_recv(ep->_cm_id, (void *)data->wr_id, (void *)data->wr_id,
                       ep->_recvMsgSize, ep->_recvMr);
        struct MessageHeader *req = (struct MessageHeader *)buffer;
        std::cout << "recv" << req->type << "\n";
        switch (req->type)
        {
        case MessageType::GET:
            ep->processGet(req);
            break;
        case MessageType::DELETE:
            ep->processDelete(req);
            break;
        case MessageType::PUT:
            ep->processPut(req);
            break;
        case MessageType::SUCCESS:
            processReplicationDone(req->id, MessageType::FAILURE, false);
            break;
        case MessageType::FAILURE:
            processReplicationDone(req->id, MessageType::FAILURE, false);
            break;
        default:
            std::ostringstream ss;
            ss << "RepRequest invalid req type" << data->opcode;
            CPPLog::LOG_ERROR(ss);
            break;
        }
        delete[] buffer;
    }
    break;
    default:
        std::ostringstream ss;
        ss << "TaskThread default opcode : " << data->opcode;
        CPPLog::LOG_INFO(ss);
        break;
    }
}

void TaskThread::processReplicationDone(uint32_t id, MessageType type, bool localWrite)
{
    std::unique_lock<std::mutex> cqEventlock(_cqEventsMutex);
    auto it = _cqEvents.find(id);
    std::cout << "size " << _cqEvents.size() << "\n";
    if (it == _cqEvents.end())
    {
        std::ostringstream ss;
        ss << "Cqevent not found " << id;
        CPPLog::LOG_ERROR(ss);
        return;
    }
    // std::cout<<"rep done\n";
    // For local write we  do not increase repdone count
    if (!localWrite)
        ++(it->second->_repDone);
    /* if there is no rep event then this condition will be false(0<0) and we can send response back to client*/
    if (it->second->_repDone < it->second->_repRequired)
    {
        // std::cout<<"event 2\n";
        return;
    }
    _cqEvents.erase(it);
    cqEventlock.unlock();

    RdmaSalEndpoint *ep = it->second->_ep;
    char *sendBuf = nullptr;
    std::unique_lock<std::mutex> lock(ep->_sendBuffersM);
    if (ep->_sendBuffers.size() == 0)
    {
        CPPLog::LOG_ERROR("No send Buffer");
        return;
    }
    sendBuf = ep->_sendBuffers.front();
    ep->_sendBuffers.pop();
    lock.unlock();
    MessageHeader *response = (MessageHeader *)sendBuf;
    response->id = id;
    response->type = type;
    int ret = rdma_post_send(ep->_cm_id, sendBuf, sendBuf, MessageHeaderSize, ep->_sendMr, 0);
    if (ret == -1)
    {
        std::ostringstream ss;
        ss << "rdma_post_send error occured errno " << errno << strerror(errno);
        CPPLog::LOG_ERROR(ss);
    }
}

void TaskThread::replicateSalRequest(char *req, uint32_t size)
{
    /* for replication we will not create a new packet we will just forward the same packet recieved from client*/
    auto repIt = _clientRepMap->begin();
    for (; repIt != _clientRepMap->end(); repIt++)
    {
        repIt->second->sendMessage(req, size);
    }

    auto serverRepIt = _serverRepMap->begin();
    for (; serverRepIt != _serverRepMap->end(); serverRepIt++)
    {
        serverRepIt->second->sendMessage(req, size);
    }
}
void TaskThread::sendInvalidation(char *req)
{
    MessageHeader *salReq = (MessageHeader *)req;
    char *buffer = new char[MessageHeaderSize + salReq->keySize];
    MessageHeader *invRequest = (MessageHeader *)(buffer);
    invRequest->type = MessageType::INVALIDATE;
    invRequest->id = salReq->id;
    invRequest->keySize = salReq->keySize;
    memcpy(buffer + MessageHeaderSize, salReq + MessageHeaderSize, salReq->keySize);
    // Send Invalidation to sal's
    auto salIt = _salMap->begin();
    for (; salIt != _salMap->end(); salIt++)
    {
        salIt->second->sendMessage(buffer, MessageHeaderSize + salReq->keySize);
    }
    delete[] buffer;
}